from typing import Any, Sequence

from llama_index.core.base.llms.types import ChatMessage, ChatResponse
from llama_index.llms.openai import OpenAI
from llama_index.llms.openai.utils import from_openai_message, to_openai_message_dicts


class CustomOpenAI(OpenAI):
    def _chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        client = self._get_client()
        message_dicts = to_openai_message_dicts(messages)
        # from pprint import pprint
        # print("22222 @@@@@@@@@@@")
        # pprint(kwargs)
        # print("################")
        # pprint(messages)
        # print("################")
        response = client.chat.completions.create(
            messages=message_dicts,
            stream=False,
            **self._get_model_kwargs(**kwargs),
        )
        openai_message = response.choices[0].message
        message = from_openai_message(openai_message)

        return ChatResponse(
            message=message,
            raw=response,
            additional_kwargs=self._get_response_token_counts(response),
        )



import logging
import os
import re
import shutil
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from langchain_community.embeddings.huggingface import HuggingFaceEmbeddings
from llama_index import (
    Document,
    ServiceContext,
    SimpleDirectoryReader,
    StorageContext,
    VectorStoreIndex,
    get_response_synthesizer,
    load_index_from_storage,
)
from llama_index.embeddings import HuggingFaceEmbedding, LangchainEmbedding
# from llama_index.indices.postprocessor import (
#     AutoPrevNextNodePostprocessor,
#     KeywordNodePostprocessor,
# )
# from llama_index.postprocessor import SimilarityPostprocessor
# from llama_index.query_engine import RetrieverQueryEngine
# from llama_index.readers.file.markdown_reader import MarkdownReader
# from llama_index.response_synthesizers import (
#     BaseSynthesizer,
#     ResponseMode,
#     get_response_synthesizer,
# )
# from llama_index.retrievers import (
#     BaseRetriever,
#     VectorIndexAutoRetriever,
#     VectorIndexRetriever,
# )
# from llama_index.schema import MetadataMode
# from llama_index.storage.storage_context import StorageContext
# from llama_index.vector_stores import ChromaVectorStore
# from llama_index.vector_stores.types import MetadataInfo, VectorStoreInfo

from qai.storage.retriever import Retriever
# from qai.core import Meta, MetaObj

# from qscraper.filters.filter import Filter
# from qscraper.processors.text_writer import TextWriter

# from qai.chat import cfg
# from qai.chat.db import Retriever
# from qai.chat.db.chroma.chroma import Chroma, ChromaConfig
# from qai.chat.layers.llm.chatbot import LLM
# from qai.chat.prompt.config import CONTEXT_TOKEN_LIMIT, HISTORY_TOKEN_LIMIT

# from qai.scraper.scrapers.scraper import WebObj

if TYPE_CHECKING:
    from chromadb import Collection

def title_format(s: str):
    ### remove everything before _com
    title = re.sub(r"^.+?_com", "", s)

    title = title.replace("_", " ")
    title = title.replace("-", " ")
    title = title.replace(".html", "")
    title = title.replace(".md", "")
    ## remove multiple spaces
    title = re.sub(r"\s+", " ", title)
    title = title.strip()
    return title


@dataclass
class LlamaAdapter(Retriever):
    index: VectorStoreIndex = None
    db_location: str = None
    index_name: str = None  ## f"{company_name}_{some_id}"
    # chroma: Chroma = None
    init_chroma: bool = True

    def __post_init__(self):
        # if "chroma_config" in self.config:
        #     self.chroma = Chroma(config=self.config["chroma_config"])
        print("Initialized LLamaIndex. self.chroma=", self.chroma)
        if self.index_name is None:
            raise Exception("index_name must be set")
        if self.init_chroma:
            chroma_collection = self.chroma.get_collection(self.index_name)
            vector_store = ChromaVectorStore(chroma_collection=chroma_collection)

            self.index = VectorStoreIndex.from_vector_store(
                vector_store=vector_store,
            )

    def get_context(
        self,
        query: str,
        k: int = 8,
    ) -> list[tuple[str, dict[str, Any]]]:
        """
        Given a query, return the context of the query.
        The context is a list of tuples (text, metadata) where text is the text of the document
        """
        nodes = self.query(query=query, k=k)
        return [(n.get_text(), n.node.source_node.metadata) for n in nodes]

    def get_index(self):
        chroma_collection = self.chroma.get_collection(self.index_name)
        vector_store = ChromaVectorStore(chroma_collection=chroma_collection)

        self.index = VectorStoreIndex.from_vector_store(
            vector_store=vector_store,
        )

    def query(self, query: str, k=8, verbose=False, **kwargs):
        print("####################")
        if self.index is None:
            raise Exception("Index is not set")
            # self.index_name = f"{gcompany}_simplified_md"
            # self.index = self.get_or_create_index(
            #     meta=gmeta,
            #     subfolder="simplified_md",
            #     needs_meta=False,
            # )
        retriever = VectorIndexRetriever(
            self.index,
            similarity_top_k=k,
        )

        node_postprocessors = [
            KeywordNodePostprocessor(
                # required_keywords=["Combinator"],
                # exclude_keywords=["Italy"],
            )
        ]
        query_engine = RetrieverQueryEngine.from_args(
            retriever,
            node_postprocessors=node_postprocessors,
            response_mode=ResponseMode.TREE_SUMMARIZE,
        )
        nodes = query_engine.retrieve(query)
        return_nodes = []
        text_hash = set()
        length = 0
        for n in nodes:
            # print(n.node.source_node, n)
            if verbose:
                print(
                    f" {query} fn ############### {n.score:.3f}",
                    n.node.source_node.metadata,
                    len(n.get_content().strip()),
                    " cur len=",
                    length,
                )
            txt = n.get_text()
            if txt in text_hash:
                continue
            text_hash.add(txt)
            length += len(txt)
            return_nodes.append(n)
        return return_nodes

    def _make_category(
        self, meta: Meta, wf: MetaObj, title: str, str_cat: str, categories: list[str]
    ):
        return "Unknown"
        category = meta.get(wf.url, {}).get("category")
        if not category or category == "Unknown" or category == "None":
            if len(categories) <= 1:
                return categories[0]
            user_msg = (
                f"Given the following markdown in triple quotes, "
                f"give the single best category among [{str_cat}] "
                "that it belongs to?\n\n"
            )
            with open(wf.full_path, "r") as f:
                t = f.read()
                user_msg += f'"""{t}"""'
            try:
                llm = LLM(config=cfg.category_model)
                print(f"Using llm to get category for {wf.url}")
                qr = llm.simple_query(user_msg)

                category = qr
                ## if quotes appear then get the category out of the quotations
                if '"' in category:
                    category = category.split('"')[1]
            except Exception as e:
                print(e)
                category = title.split(" ")[0]
            meta.save()
        return category

    def _get_or_create_index(
        self,
        meta: Meta,
        # website_dir: str,
        subfolder: str = None,
        overwrite: bool = False,
        categories: list[str] = None,
        needs_meta: bool = True,
    ) -> "LlamaAdapter":
        db_location = self.db_location
        col = self.chroma.get_or_create_collection(self.index_name)
        if col.count() > 0:
            chroma_collection = self.chroma.get_or_create_collection(self.index_name)
            vector_store = ChromaVectorStore(chroma_collection=chroma_collection)
            storage_context = StorageContext.from_defaults(vector_store=vector_store)
            self.index = VectorStoreIndex.from_vector_store(
                vector_store=vector_store,
            )
            return self
        print(
            f"Creating index at {db_location} from {meta.get_directory()} with subfolder {subfolder}"
        )
        if not categories:
            categories = ["Unknown"]
        str_cat = ", ".join(categories)

        storage_context = None

        if db_location:
            db_location = os.path.expanduser(db_location)
            if overwrite and os.path.exists(db_location):
                shutil.rmtree(db_location)
            os.makedirs(db_location, exist_ok=True)

        elif os.path.exists(f"{db_location}/docstore.json"):
            print(f"Loading index from {db_location}")
            # rebuild storage context
            storage_context = StorageContext.from_defaults(persist_dir=db_location)
            # # load index
            # index = load_index_from_storage(storage_context)
            logging.info(f"Loaded index from {db_location}")
            return LlamaAdapter(index=index)
        elif self.chroma is not None:
            print(f"Creating index with chroma. {self.chroma.chroma_client} ")
            chroma_collection = self.chroma.get_or_create_collection(self.index_name)
            vector_store = ChromaVectorStore(chroma_collection=chroma_collection)
            storage_context = StorageContext.from_defaults(vector_store=vector_store)
            # self.index = VectorStoreIndex.from_vector_store(
            #     vector_store=vector_store,
            # )

        documents = []
        meta_link = {}

        def meta_fn(filename):
            if filename in meta_link:
                return meta_link[filename]
            return {}

        ## Process files
        for wf in meta.get_files_meta(subfolder=subfolder, cls=MetaObj):
            print(f"subfolders={subfolder} WF {wf}")
            title = title_format(wf.uri.split("/")[-1])
            if not title:
                title = "Home Page"
            try:
                category = str(wf.category)

                if wf.category == "Unknown":
                    self._make_category(meta, wf, title, str_cat, categories)
            except Exception as e:
                category = self._make_category(meta, wf, title, str_cat, categories)

            wf_meta = {
                "url": wf.uri,
                "title": title,
                "category": category.lower(),
            }
            meta_link[str(wf.uri)] = wf_meta

            # print("     ", wf_meta, wf.full_path)
            ### TODO add metadata to the document
            # meta.meta[wf.url] = wf_meta
        d = meta.get_directory(subfolder=subfolder)
        documents = SimpleDirectoryReader(d, file_metadata=meta_fn, filename_as_id=True).load_data()

        meta.save()
        index = VectorStoreIndex.from_documents(
            documents=documents,
            storage_context=storage_context,
        )
        if self.chroma is None:
            index.storage_context.persist(persist_dir=db_location)

        self.index = index
        return self



import json
from collections import defaultdict
from dataclasses import dataclass, field

import boto3
from pi_log import logs

from qai.storage import cfg

log = logs.getLogger(__name__)

s3 = boto3.client("s3")

bucket = boto3.resource("s3").Bucket("suparev-chat-storage")


def load_s3(key):
    return bucket.Object(key=key).get()["Body"].read().decode("utf-8")


def load_s3_json(key):
    return json.loads(load_s3(key))


class AWSCompanyIds:
    def __init__(self):
        self.website_2_id = load_s3_json(key=cfg.aws.company_ids_file)

        self.id_2_website = {v: k for k, v in self.website_2_id.items()}


@dataclass
class AWSHistory:
    company_id: str
    history: dict[str, list[dict[str, str]]] = field(
        default_factory=dict
    )  # user id to list of messages
    seen: set[str] = None

    def _make_file_name(self, user_id: str) -> str:
        file_name = cfg.aws.history_file
        file_name = file_name.replace("<company_id>", self.company_id)
        file_name = file_name.replace("<user_id>", user_id)
        return file_name

    def get_history(self, user_id: str, ignore_warnings: bool = False) -> list[dict[str, str]]:
        if self.history is None:
            self.history = defaultdict(list)
        if user_id in self.history:
            return self.history[user_id]
        file_name = self._make_file_name(user_id=user_id)
        try:
            self.history[user_id] = load_s3_json(key=file_name)
        except Exception as e:
            if not ignore_warnings:
                log.error(f"Failing to load history for {self.company_id}, {user_id}")
                log.error(e)
            self.history[user_id] = []
        return self.history[user_id]

    def __post_init__(self):
        if not self.company_id:
            raise ValueError("Must provide company id")

    def add_history(
        self, user_id: str, messages: dict[str, str] | list[dict[str, str]], ignore_duplicates=True
    ):
        if isinstance(messages, dict):
            messages = [messages]
        if user_id not in self.history:
            self.history[user_id] = []
        if ignore_duplicates:
            if len(self.history[user_id]) > 0:
                last_message = self.history[user_id][-1]
                if last_message == messages[0]:
                    return
        self.history[user_id].extend(messages)
        self.history[user_id] = self.history[user_id][-100:]

    def save(self, user_id: str):
        file_name = self._make_file_name(user_id=user_id)
        hist = self.history[user_id]
        s3.put_object(Body=json.dumps(hist), Bucket=bucket.name, Key=file_name)

    def delete(self, user_id: str, ignore_missing: bool = True):
        file_name = self._make_file_name(user_id=user_id)
        if ignore_missing:
            try:
                s3.delete_object(Bucket=bucket.name, Key=file_name)
            except:
                pass
        else:
            s3.delete_object(Bucket=bucket.name, Key=file_name)

    def exists(self, user_id: str) -> bool:
        file_name = self._make_file_name(user_id=user_id)
        try:
            load_s3(key=file_name)
            return True
        except:
            return False



from pathlib import Path
import os
import glob
import boto3

def download_dir(bucket_name, s3_folder, local_dir=None):
    """
    From: https://stackoverflow.com/questions/49772151/download-a-folder-from-s3-using-boto3
    Download the contents of a folder directory
    Args:
        bucket_name: the name of the s3 bucket
        s3_folder: the folder path in the s3 bucket
        local_dir: a relative or absolute directory path in the local file system
    """
    if not local_dir is None:
        local_dir = os.path.expanduser(local_dir)
    s3 = boto3.resource("s3")
    bucket = s3.Bucket(bucket_name)
    for obj in bucket.objects.filter(Prefix=s3_folder):
        target = obj.key if local_dir is None \
            else os.path.join(local_dir, os.path.relpath(obj.key, s3_folder))
        if not os.path.exists(os.path.dirname(target)):
            os.makedirs(os.path.dirname(target))
        if obj.key[-1] == '/':
            continue
        bucket.download_file(obj.key, target)

def upload_dir(localDir, awsInitDir, bucketName, tag="*", prefix="/"):
    """
    from https://stackoverflow.com/questions/56426471/upload-folder-with-sub-folders-and-files-on-s3-using-python
    from current working directory, upload a 'localDir' with all its subcontents (files and subdirectories...)
    to a aws bucket
    Parameters
    ----------
    localDir :   localDirectory to be uploaded, with respect to current working directory
    awsInitDir : prefix 'directory' in aws
    bucketName : bucket in aws
    tag :        tag to select files, like *png
                NOTE: if you use tag it must be given like --tag '*txt', in some quotation marks... for argparse
    prefix :     to remove initial '/' from file names

    Returns
    -------
    None
    """
    s3 = boto3.resource("s3")
    cwd = str(Path.cwd())
    p = Path(os.path.join(Path.cwd(), localDir))
    remove_prefix = os.path.join(*p.parts[:-1])
    print(f"cwd={cwd}  p={p}  remove_prefix={remove_prefix}")

    mydirs = list(p.glob("**"))
    for mydir in mydirs:
        fileNames = glob.glob(os.path.join(mydir, tag))
        fileNames = [f for f in fileNames if not Path(f).is_dir()]
        rows = len(fileNames)
        for i, fileName in enumerate(fileNames):
            fileName = str(fileName).replace(cwd, "")
            if fileName.startswith(prefix):  # only modify the text if it starts with the prefix
                fileName = fileName.replace(prefix, "", 1)  # remove one instance of prefix
            if not fileName.startswith("/"):
                fileName = "/" + fileName

            awsPath = os.path.join(awsInitDir, str(fileName))
            awsPath = awsPath.replace(remove_prefix, "", 1)
            print(f"{i:03d}/{rows}: fileName {fileName}  aws={awsPath}")
            s3.meta.client.upload_file(fileName, bucketName, awsPath)

## list the contents of the bucket
# s3 = boto3.client("s3")
# response = s3.list_objects_v2(Bucket=bucket)
# for obj in response.get("Contents", []):
#     print(obj["Key"])



from dataclasses import dataclass, field

from qai.storage.aws.aws_history import AWSHistory


@dataclass
class History:
    chat_history: dict[str, "CompanyHistory"] = field(default_factory=dict)

    def get_history(self, company_id: str, user_id: str) -> list[dict[str, str]]:
        if company_id not in self.chat_history:
            self.chat_history[company_id] = CompanyHistory(company_id=company_id)
        company_history = self.chat_history[company_id]
        return company_history.get_history(user_id=user_id)

    def add_history(
        self,
        company_id: str,
        user_id: str,
        messages: dict[str, str] | list[dict[str, str]],
        save: bool = False,
        ignore_duplicates: bool = True,
    ):
        if company_id not in self.chat_history:
            self.chat_history[company_id] = CompanyHistory(company_id=company_id)
        company_history = self.chat_history[company_id]
        company_history.add_history(
            user_id=user_id, messages=messages, ignore_duplicates=ignore_duplicates
        )
        if save:
            company_history.save(user_id=user_id)

    def get_add_history(
        self,
        company_id: str,
        user_id: str,
        messages: dict[str, str] | list[dict[str, str]],
        save: bool = False,
        ignore_duplicates: bool = True,
    ) -> list[dict[str, str]]:
        self.add_history(
            company_id=company_id,
            user_id=user_id,
            messages=messages,
            save=save,
            ignore_duplicates=ignore_duplicates,
        )
        return self.get_history(company_id=company_id, user_id=user_id)


@dataclass
class CompanyHistory(AWSHistory):
    pass
